'''
Test of strains of LaminateStrength
===================================
 
Test LaminateStrength class using examples from reference 1

References
----------
1 - https://ntrs.nasa.gov/api/citations/19950009349/downloads/19950009349.pdf
'''

import pytest
import numpy as np
import pandas as pd

from composipy import OrthotropicMaterial, LaminateProperty, LaminateStrength


@pytest.fixture
def laminateconfig():
    '''
    Material configuration from Example 1 in page 31.
    Laminate configuration from Example 3 in page 46
    This laminate is used in all tests
    '''
    E1 = 2.01e7 #lb/in2
    E2 = 1.301e6 #lb/in2
    v12 = 0.3
    G12 = 1.001e6 #lb/in2
    t = 0.005 #in
    material = OrthotropicMaterial(E1, E2, v12, G12, t)
    stacking = [0, 45, 45, 0]
    return LaminateProperty(stacking, material)


@pytest.fixture
def example_3_setup(laminateconfig):
    '''
    Example 3 setup in page 46
    '''
    Nxx = 500 #lb/in
    return LaminateStrength(dproperty=laminateconfig, Nxx=Nxx)


@pytest.fixture
def example_6_setup(laminateconfig):
    '''
    Example 6 setup in page 56
    '''
    Mxx = 5 #lb/in
    return LaminateStrength(dproperty=laminateconfig, Mxx=Mxx)


def test_example_3_midplane(example_3_setup):
    '''This test check the stains at the midplane'''
    calculated_strains = example_3_setup.epsilon0()
    reference_result = np.array([0.00221, -0.0007, -0.00115, 0, 0, 0])
    np.testing.assert_allclose(calculated_strains, reference_result, rtol=1e-2, atol=1e-4)


def test_example_3_strain_45(example_3_setup):
    '''This test check the stains for the 45° plies in material direction'''
    df_strains = example_3_setup.calculate_strain()
    epsilon1 = df_strains.iloc[2]['epsilon1']
    epsilon2 = df_strains.iloc[2]['epsilon2']
    gamma12 = df_strains.iloc[2]['gamma12']
    assert np.isclose(epsilon1, 0.00018, rtol=1e-3, atol=1e-4)
    assert np.isclose(epsilon2, 0.00133, rtol=1e-3, atol=1e-4)
    assert np.isclose(gamma12, -0.00291, rtol=1e-3, atol=1e-4)


def test_example_3_stress_45(example_3_setup):
    '''This test check the stresses for the 45° plies in material direction'''
    df_stresses = example_3_setup.calculate_stress()
    sigma1 = df_stresses.iloc[2]['sigma1']
    sigma2 = df_stresses.iloc[2]['sigma2']
    tau12 = df_stresses.iloc[2]['tau12']
    assert np.isclose(sigma1, 4146, rtol=1e-3, atol=100)
    assert np.isclose(sigma2, 1812, rtol=1e-3, atol=100)
    assert np.isclose(tau12, -2913, rtol=1e-3, atol=100)


def test_example_6_midplane(example_6_setup):
    '''This test check the stains at the midplane eq E6.1'''
    calculated_strains = example_6_setup.epsilon0()
    reference_result = np.array([0, 0, 0, 0.418, -0.165, -0.0975])
    np.testing.assert_allclose(calculated_strains, reference_result, rtol=1e-2, atol=1e-4)


def test_example_6_strain_top_ply4(example_6_setup):
    '''
    This test check the stains at the top of ply 4 in laminate direction    
    Please note that ply 4 in composipy is ply 1 in NASA reference document.
    Composipy table starts from the bottom to the top.   
    '''
    df_strains = example_6_setup.calculate_strain()
    epsilonx = df_strains.loc[7, 'epsilonx']
    epsilony = df_strains.loc[7, 'epsilony']
    gammaxy = df_strains.loc[7, 'gammaxy']
    assert np.isclose(epsilonx, 0.00418, rtol=1e-3, atol=1e-4)
    assert np.isclose(epsilony, -0.00165, rtol=1e-3, atol=1e-4)
    assert np.isclose(gammaxy, -0.000975, rtol=1e-3, atol=1e-4)


def test_example_6_strain_top_ply3(example_6_setup):
    '''
    This test check the stains at the top of ply 3 in material direction
    Please note that ply 3 in composipy is ply 2 in NASA reference document.
    Composipy table starts from the bottom to the top.   
    '''

    df_strains = example_6_setup.calculate_strain()
    epsilon1 = df_strains.loc[5, 'epsilon1']
    epsilon2 = df_strains.loc[5, 'epsilon2']
    gamma12 = df_strains.loc[5, 'gamma12']
    assert np.isclose(epsilon1, 0.00039, rtol=1e-3, atol=1e-4)
    assert np.isclose(epsilon2, 0.00088, rtol=1e-3, atol=1e-4)
    assert np.isclose(gamma12, -0.00292, rtol=1e-3, atol=1e-4)


def test_example_6_stress_bot_ply2(example_6_setup):
    '''
    This test check the stresses for the ply 3 in material direction E6.12
    Please note that ply 2 in composipy is ply 3 in NASA reference document.
    Composipy table starts from the bottom to the top.  
    '''
    df_stresses = example_6_setup.calculate_stress()
    sigma1 = df_stresses.loc[2, 'sigma1']
    sigma2 = df_stresses.loc[2, 'sigma2']
    tau12 = df_stresses.loc[2, 'tau12']
    assert np.isclose(sigma1, -8094, rtol=1e-3, atol=100)
    assert np.isclose(sigma2, -1296, rtol=1e-3, atol=100)
    assert np.isclose(tau12, 2923, rtol=1e-3, atol=100)


import numpy as np
import matplotlib.pyplot as plt

from scipy.optimize import NonlinearConstraint, Bounds, minimize, LinearConstraint
from scipy.sparse.linalg import ArpackError

from .utils import Ncr_from_lp, normalize_critical_load, penalty_g1, penalty_g2, check_loads, natural_constraint_g, plot_optimization


__all__ = ['maximize_buckling_load']


def maximize_buckling_load(a, b, T,
                   E1, E2, v12, G12,
                   Nxx=0, Nyy=0, Nxy=0,
                   m=7, n=7,
                   panel_constraint='PINNED',
                   options=None,
                   tol=None,
                   plot=False,
                   points_to_plot=30,
                   penalty=True
                   ):
    '''
    Parameters
    ----------
    a : float
        Plate dimension a
    b : float
        Plate dimension b
    E1 : float
        Young Modulus direction 1
    E2 : float
        Young Modulus direction 2
    v12 : float
        Poisson ratio
    G12 : float
        Shear Modulus
    Nxx : float, default 0
        Normal load in x direction
    Nyy : float, default 0
        Normal load in y direction
    Nxy : float, default 0
        Normal load in xy direction
    m : float, dafault 7
        Size of shape function along x axis
    n : float, default 0
        Size of shape function along y axis
    panel_constraint : composipy constraints format
        Plate boundary constraints
    options : scipyminimize optiions
    tol : scipy minimize tol
    plot : bool, default False
        Plot optimization function
    points_to_plot : float, default 30
      Number of points to plot
    penalty : bool, default True
      Applies boundary conditions into the feasible region.
      If penalty is True, the boundary condition is the triangle the delimits 0, 45, 90°
      If penalty is False, the boundary conditions is the parabola that defines the natural constraints.

    Returns
    ------- 
    res : scipy minimize result
    '''

    check_loads(Nxx, Nyy, Nxy)

    # Normalizing loads
    Nxx_norm, Nyy_norm, Nxy_norm, max_load = normalize_critical_load(Nxx, Nyy, Nxy)

    # Critical load error handling
    def critical_load(y0):
        xi1, xi3 = y0
        try:
            eig = Ncr_from_lp(a, b, T, m, n, xi1, xi3, E1, E2, v12, G12, Nxx_norm, Nyy_norm, Nxy_norm, constraints=panel_constraint)
        except ArpackError: #cause if xi1 and xi3 chosen by optimizer is out of the bounds, it may crash
            eig = -1
        return (-1) * eig #negative so the solver can maximize instead of minimize
    
    # Constraints and boundaries   
    b1 = ([-1.0, 1.0], [-1.0, 1.0])
    x0 = [0.0, 0.0]
    if penalty:
        c2 = NonlinearConstraint(penalty_g1, -0.0001, 10000)
        c3 = NonlinearConstraint(penalty_g2, -0.0001, 10000)
        res = minimize(critical_load, x0, method='SLSQP', constraints=[c2, c3], bounds=b1, options=options, tol=tol)
    elif not penalty:
        c = NonlinearConstraint(natural_constraint_g, -0.0001, 10000)
        res = minimize(critical_load, x0, method='SLSQP', constraints=[c], bounds=b1, options=options, tol=tol)
 

    if plot:
        plot_optimization(a, b, T, m, n, E1, E2, v12,
                       G12, Nxx, Nyy, Nxy, panel_constraint, points_to_plot, res, penalty)

    return res



import numpy as np

from composipy import OrthotropicMaterial, LaminateProperty, PlateStructure
from scipy.optimize import NonlinearConstraint, Bounds, minimize, LinearConstraint
from scipy.sparse.linalg import ArpackError

from .utils import Ncr_from_lp, normalize_critical_load, penalty_g1, penalty_g2, check_loads


__all__ = ['minimize_panel_weight']


def _objective_function(y0):
    T, *_ = y0 #(T, xi1, xi3)
    return T**3


def minimize_panel_weight(a, b,
                   E1, E2, v12, G12,
                   Nxx=0, Nyy=0, Nxy=0,
                   m=7, n=7,
                   x0 = [0.1, 0.0, 0.0],
                   panel_constraint='PINNED',
                   options=None,
                   tol=None,
                   ):
    '''
    Parameters
    ----------
    a : float
        Plate dimension a
    b : float
        Plate dimension b
    E1 : float
        Young Modulus direction 1
    E2 : float
        Young Modulus direction 2
    v12 : float
        Poisson ratio
    G12 : float
        Shear Modulus
    Nxx : float, default 0
        Normal load in x direction
    Nyy : float, default 0
        Normal load in y direction
    Nxy : float, default 0
        Normal load in xy direction
    m : float, dafault 7
        Size of shape function along x axis
    n : float, default 0
        Size of shape function along y axis
    x0 : list, default [0.1, 0.0, 0.0]
        Initial value
    panel_constraint : composipy constraints format
        Plate boundary constraints
    options : scipyminimize optiions
    tol : scipy minimize tol
    plot : bool, default False
        Plot optimization function
    points_to_plot : float, default 30
      Number of points to plot

    Returns
    ------- 
    res : scipy minimize result
    '''

    check_loads(Nxx, Nyy, Nxy)

    # Normalizing loads
    Nxx_norm, Nyy_norm, Nxy_norm, max_load = normalize_critical_load(Nxx, Nyy, Nxy)

    # Critical load error handling
    def critical_load(y0):
        T, xi1, xi3 = y0
        try:
            eig = Ncr_from_lp(a, b, T, m, n, xi1, xi3, E1, E2, v12, G12, Nxx_norm, Nyy_norm, Nxy_norm, constraints=panel_constraint)
        except ArpackError: #cause if xi1 and xi3 chosen by optimizer is out of the bounds, it may crash
            eig = -1
        return eig
    
    # Constraints and boundaries
    c1 = NonlinearConstraint(critical_load, max_load, np.inf)
    c2 = NonlinearConstraint(penalty_g1, -0.0001, 10000)
    c3 = NonlinearConstraint(penalty_g2, -0.0001, 10000)
    b1 = ([0.001, 1000000], [-1.0, 1.0], [-1.0, 1.0])
   
    res = minimize(_objective_function, x0, method='SLSQP',
                   constraints=[c1, c2, c3], bounds=b1, options=options, tol=tol)

    return res



import pytest
import numpy as np

from composipy import OrthotropicMaterial, LaminateProperty


@pytest.fixture
def stacking_plies():
    E11 = 120000
    E22 = 6500
    nu12 = 0.07
    G12 = 4550
    t = 0.21

    laminaprop = (E11, E22, nu12, G12, G12, G12)
    stack = [45, -45, 0, 90, 45, 0, 0]
    stack += stack[::-1]

    mat1 = OrthotropicMaterial(E11, E22, nu12, G12, t)
   
    return stack, mat1


def test_xiA(stacking_plies):
    stacking, plies = stacking_plies
    xiA_reference = np.array([
        0.2857142857142857,
        0.14285714285714285,
        0.14285714285714285,
        -1.7494954273533614e-17
    ]) # calculated with composites
    laminate = LaminateProperty(stacking, plies)
    np.testing.assert_almost_equal(laminate.xiA, xiA_reference)


def test_xiD(stacking_plies):
    stacking, plies = stacking_plies
    xiD_reference = np.array([
        0.09329446064139943,
        0.16034985422740528,
        -0.38192419825072876,
        -6.783757779533445e-18
    ]) # calculated with composites
    laminate = LaminateProperty(stacking, plies)
    np.testing.assert_almost_equal(laminate.xiD, xiD_reference)


import numpy as np
import pandas as pd

#TODO:
# add more advanced strength criteria

__all__ = ['LaminateStrength']


class LaminateStrength():
    '''
    Creates a LaminateStrength object to evaluate strength.
    The class is capable of calculating strain and stress at the midplane, as well
    as ply-by-ply at the top and bottom, both in the laminate and material directions.

    Parameters
    ----------
    dproperty : LaminateProperty
        A laminate property object
    Nxx : float, int, optional, default 0
        Membrane load in x direction.
    Nyy : float, int, optional, default 0
        Membrane load in y direction.    
    Nxy : float, int, optional, default 0
        Membrane load in xy direction.
    Mxx : float, int, optional, default 0
        Moment in x direction.
    Myy : float, int, optional, default 0
        Moment in y direction.
    Mxy : float, int, optional, default 0
        Moment in xy direction.
    '''

    def __init__(
            self, 
            dproperty, 
            Nxx=0,
            Nyy=0,
            Nxy=0,
            Mxx=0,
            Myy=0,
            Mxy=0):
        self.dproperty = dproperty
        self.Nxx = Nxx
        self.Nyy = Nyy
        self.Nxy = Nxy
        self.Mxx = Mxx
        self.Myy = Myy
        self.Mxy = Mxy
        self._positions = None #used for plotting.


    def epsilon0(self): #ok
        '''
        Calculates the strains of the laminate
        
        Returns
        -------
        epsilon0 : numpy ndarray
            [epsilonx0, epsilony0, epsilonxy0, kappax0, kappay0, kappaxy0]
        '''

        ABD = self.dproperty.ABD
        abd = np.linalg.inv(ABD) #equivalent relations on pg 153 of Daniel
        N = np.array([self.Nxx, self.Nyy, self.Nxy, self.Mxx, self.Myy, self.Mxy])
        return abd @ N


    def _epsilonk(self): #ok
        '''
        Calculates the strains for each lamina
        Returns
        -------
        epsilonk: list
            [epsilon1, epsilon2, ..., epsilonk], with epsilon as np.ndarray()
            For each epsilon we have
            epsilonk = np.ndarray([epsilonx, epsilony, gammas])k, coordinate directions
            For reference refer to page 145 of Daniel equation 5.8
        '''
        nplies = len(self.dproperty.stacking)
        z = self.dproperty.z_position
        zmid = [(z[i], z[i+1]) #tuple with top and bot
               for i in range(nplies)]
        epsilon0 = self.epsilon0()
        epsilon = np.array([epsilon0[0],
                            epsilon0[1],
                            epsilon0[2],])
        kappa = np.array([epsilon0[3],
                          epsilon0[4],
                          epsilon0[5],])        
        epsilonk = []
        for htop, hbot in zmid:
            epsilonk.append(
                (epsilon + htop * kappa,
                 epsilon + hbot * kappa)
            )
        return epsilonk


    def _stressk(self):
        '''
        Calculates the strains for each lamina
        Returns
        -------
        epsilonk: list
            [epsilon1, epsilon2, ..., epsilonk], with epsilon as np.ndarray()
            For each epsilon we have
            epsilonk = np.ndarray([epsilonx, epsilony, gammas])k, coordinate directions
            For reference refer to page 145 of Daniel equation 5.8
        '''
        Qk = self.dproperty.Q_layup
        epsilonk = self._epsilonk()
        
        stressk = []
        for Q, epsilon in zip(Qk, epsilonk):
            epsilontop, epsilonbot = epsilon
            stressk.append(
                (Q @ epsilontop,
                Q @ epsilonbot)
            )
        return stressk

    def _epsilonk_123(self):

        stacking = self.dproperty.stacking
        epsilonk = self._epsilonk()
        
        epsilonk_123 = []
        for theta, epsilon in zip(stacking, epsilonk):
            epsilontop, epsilonbot = epsilon           
            c = np.cos(theta*np.pi/180)
            s = np.sin(theta*np.pi/180)
            epsilontop[2] /= 2 # tensorial shear strain (see nasa pg 50)
            epsilonbot[2] /= 2

            T = np.array([
                [c**2, s**2, 2*c*s],
                [s**2, c**2, -2*c*s],
                [-c*s, c*s, c**2-s**2]
                ])

            cur_epsilontop = T @ epsilontop
            cur_epsilonbot = T @ epsilonbot
            cur_epsilontop[2] *= 2
            cur_epsilonbot[2] *= 2
            epsilonk_123.append((cur_epsilontop, cur_epsilonbot)) # engineering shear strain (see nasa pg 50)
        
        return epsilonk_123


    def _stressk_123(self):

        stacking = self.dproperty.stacking
        stressk = self._stressk()
        
        stressk_123 = []
        for theta, stress in zip(stacking, stressk):
            stresstop, stressbot = stress   
            c = np.cos(theta*np.pi/180)
            s = np.sin(theta*np.pi/180)
            #stresstop[2] /= 2 # tensorial shear strain (see nasa pg 50)
            #stressbot[2] /=

            T = np.array([
                [c**2, s**2, 2*c*s],
                [s**2, c**2, -2*c*s],
                [-c*s, c*s, c**2-s**2]
                ])

            cur_stresstop = T @ stresstop
            cur_stressbot = T @ stressbot
            #cur_stresstop[2] *= 2
            #cur_stressbot[2] *= 2
            stressk_123.append((cur_stresstop, cur_stressbot)) 
        
        return stressk_123

    def _allowablek_123(self):
        '''
        Finds the allowable stresses for each lamina
        Returns
        -------
        allowablek_123: list
            [allowable1, allowable2, ..., allowablek], with allowable as np.ndarray()
            For each allowable we have
            allowablek = np.ndarray([T1, C1, T2, C2, S])k
        '''
        laminate = self.dproperty.layup

        allowablek_123 = []
        for layer in laminate:
            angle,ply = layer
            maxstresses = np.array([ply.t1,
                                    ply.c1,
                                    ply.t2,
                                    ply.c2,
                                    ply.s,])
            allowablek_123.append(maxstresses)

        return allowablek_123

    def calculate_strain(self):
        '''
        Calculates strain ply by at laminate direction and material direction.

        Returns
        -------
        strains : pd.Dataframe
            ply by ply strains in plate direction and material direction       

        Note
        ----
        The sequence of the DataFrame starts from the BOTTOM OF THE LAYUP to the TOP  OF THE LAYUP.
        When defining the laminate, the first element of the list corresponds to the bottom-most layer. This is especially important for non-symmetric laminates.
        '''
        epsilonk = self._epsilonk()
        epsilonk_123 = self._epsilonk_123()
        stacking = self.dproperty.stacking
        z = self.dproperty.z_position

        cur_ply = 1
        data = {}
        data['ply'] = []
        data['z'] = [] # so the user can check and plot ply position
        data['position'] = []
        data['angle'] = []       
        data['epsilonx'] = []
        data['epsilonx'] = []
        data['epsilony'] = []
        data['epsilony'] = []
        data['gammaxy'] = []
        data['gammaxy'] = []
        data['epsilon1'] = []
        data['epsilon1'] = []
        data['epsilon2'] = []
        data['epsilon2'] = []
        data['gamma12'] = []
        data['gamma12'] = []
        for i, epsk_eps123_stck in  enumerate(zip(epsilonk, epsilonk_123, stacking)):
            epsilon, epsilon123, theta = epsk_eps123_stck
            epsilontop, epsilonbot = epsilon
            epsilon123top, epsilon123bot = epsilon123

            data['ply'].append(cur_ply)
            data['ply'].append(cur_ply)
            data['z'].append(z[i])
            data['z'].append(z[i+1])
            data['position'].append('bot')
            data['position'].append('top')
            data['angle'].append(theta)
            data['angle'].append(theta)            
            data['epsilonx'].append(epsilontop[0]) #plate direction
            data['epsilonx'].append(epsilonbot[0])
            data['epsilony'].append(epsilontop[1])
            data['epsilony'].append(epsilonbot[1])
            data['gammaxy'].append(epsilontop[2])
            data['gammaxy'].append(epsilonbot[2])
            data['epsilon1'].append(epsilon123top[0]) #material direction
            data['epsilon1'].append(epsilon123bot[0])
            data['epsilon2'].append(epsilon123top[1])
            data['epsilon2'].append(epsilon123bot[1])
            data['gamma12'].append(epsilon123top[2])
            data['gamma12'].append(epsilon123bot[2])
            cur_ply += 1
        pd.set_option('display.precision', 2)
        return pd.DataFrame(data)


    def calculate_stress(self):
        '''
        Calculates stress ply by at laminate direction and material direction.

        Returns
        -------
        stress : pd.Dataframe
            ply by ply stress in plate direction and material direction       

        Note
        ----
        The sequence of the DataFrame starts from the BOTTOM OF THE LAYUP to the TOP  OF THE LAYUP.
        When defining the laminate, the first element of the list corresponds to the bottom-most layer. This is especially important for non-symmetric laminates.
        '''

        stressk = self._stressk()
        stressk_123 = self._stressk_123()
        stacking = self.dproperty.stacking
        z = self.dproperty.z_position

        cur_ply = 1
        data = {}
        data['ply'] = []
        data['z'] = []
        data['position'] = []
        data['angle'] = []
        data['sigmax'] = []
        data['sigmax'] = []
        data['sigmay'] = []
        data['sigmay'] = []
        data['tauxy'] = []
        data['tauxy'] = []
        data['sigma1'] = []
        data['sigma1'] = []
        data['sigma2'] = []
        data['sigma2'] = []
        data['tau12'] = []
        data['tau12'] = []
        for i, sig_sig123_theta in enumerate(zip(stressk, stressk_123, stacking)):
            sigma, sigma123, theta = sig_sig123_theta
            sigmatop, sigmabot = sigma
            sigma123top, sigma123bot = sigma123

            data['ply'].append(cur_ply)
            data['ply'].append(cur_ply)
            data['z'].append(z[i])
            data['z'].append(z[i+1])
            data['position'].append('bot')
            data['position'].append('top')
            data['angle'].append(theta)
            data['angle'].append(theta)
            data['sigmax'].append(sigmatop[0]) #plate direction
            data['sigmax'].append(sigmabot[0])
            data['sigmay'].append(sigmatop[1])
            data['sigmay'].append(sigmabot[1])
            data['tauxy'].append(sigmatop[2])
            data['tauxy'].append(sigmabot[2])
            data['sigma1'].append(sigma123top[0]) #material direction
            data['sigma1'].append(sigma123bot[0])
            data['sigma2'].append(sigma123top[1])
            data['sigma2'].append(sigma123bot[1])
            data['tau12'].append(sigma123top[2])
            data['tau12'].append(sigma123bot[2])
            cur_ply += 1
        pd.set_option('display.precision', 2)
        return pd.DataFrame(data)

    def _calculate_maxstressallowable(self):
        '''
        Calculates max stress allowable for each ply.

        Returns
        -------
        allowable : pd.Dataframe
            ply by ply max stress allowables     

        Note
        ----
        The sequence of the DataFrame starts from the BOTTOM OF THE LAYUP to the TOP  OF THE LAYUP.
        When defining the laminate, the first element of the list corresponds to the bottom-most layer. This is especially important for non-symmetric laminates.
        '''

        allowablek_123 = self._allowablek_123()
        stacking = self.dproperty.stacking

        cur_ply = 1
        data = {}
        data['ply'] = []
        data['position'] = []
        data['angle'] = []
        data['angle'] = []
        data['t1'] = []
        data['t1'] = []
        data['c1'] = []
        data['c1'] = []
        data['t2'] = []
        data['t2'] = []
        data['c2'] = []
        data['c2'] = []
        data['s'] = []
        data['s'] = []
        for allowable123, theta in zip(allowablek_123, stacking):
            data['ply'].append(cur_ply)
            data['ply'].append(cur_ply)
            data['position'].append('top')
            data['position'].append('bot')
            data['angle'].append(theta)
            data['angle'].append(theta)
            data['t1'].append(allowable123[0])
            data['t1'].append(allowable123[0])
            data['c1'].append(allowable123[1])
            data['c1'].append(allowable123[1])
            data['t2'].append(allowable123[2])
            data['t2'].append(allowable123[2])
            data['c2'].append(allowable123[3])
            data['c2'].append(allowable123[3])
            data['s'].append(allowable123[4])
            data['s'].append(allowable123[4])
            cur_ply += 1
        pd.set_option('display.precision', 2)
        return pd.DataFrame(data)

    def calculate_maxstressmargin(self):
        '''
        Calculates margin of safety for each ply according to the max stress criteria.

        Returns
        -------
        margin : pd.Dataframe
            ply by ply max stress margin of safety     

        Note
        ----
        The sequence of the DataFrame starts from the BOTTOM OF THE LAYUP to the TOP  OF THE LAYUP.
        When defining the laminate, the first element of the list corresponds to the bottom-most layer. This is especially important for non-symmetric laminates.
        '''
        df_allowable = self._calculate_maxstressallowable()
        df_stress = self.calculate_stress()

        margin_t1 = df_allowable['t1'] / df_stress['sigma1'].mask(df_stress['sigma1'] < 0, 0) - 1 # replace negative stress with zero to yield infinite margins in tension
        margin_c1 = df_allowable['c1'] / df_stress['sigma1'].mask(df_stress['sigma1'] > 0, 0).abs() - 1 # replace positive stress with zero to yield infinite margins in compression. Also flip stress sign since allowables are all positive
        margin_t2 = df_allowable['t2'] / df_stress['sigma2'].mask(df_stress['sigma2'] < 0, 0) - 1
        margin_c2 = df_allowable['c2'] / df_stress['sigma2'].mask(df_stress['sigma2'] > 0, 0).abs() - 1
        margin_s = df_allowable['s'] / df_stress['tau12'].abs() - 1

        df_margin = pd.DataFrame({})
        df_margin['ply'] = df_allowable['ply']
        df_margin['position'] = df_allowable['position']
        df_margin['angle'] = df_allowable['angle']
        df_margin['margin_t1'] = margin_t1
        df_margin['margin_c1'] = margin_c1
        df_margin['margin_t2'] = margin_t2
        df_margin['margin_c2'] = margin_c2
        df_margin['margin_s'] = margin_s
        df_margin['critical'] = df_margin[['margin_t1','margin_c1','margin_t2','margin_c2','margin_s']].idxmin(axis=1) # find the critical failure mode of each ply
        pd.set_option('display.precision', 2)

        return df_margin


'''Utilities functions to be used in optimization'''

import warnings

import numpy as np
import matplotlib.pyplot as plt

from composipy import OrthotropicMaterial, LaminateProperty, PlateStructure


def Ncr_from_lp(a, b, T, m, n, xi1, xi3, E1, E2, v12, G12, Nxx, Nyy, Nxy, constraints):
    '''
    Calculates critical buckling load based in lamination parameters, geometry and constraints
    '''
    
    stacking = {
    'xiD': [xi1, 0.0, xi3, 0.0],
     'T': T}

    mat1 = OrthotropicMaterial(E1, E2, v12, G12, thickness=0.1)
    laminate1 = LaminateProperty(stacking, mat1)
    plate1 = PlateStructure(dproperty=laminate1, a=a, b=b, constraints=constraints, Nxx=Nxx, Nyy=Nyy, Nxy=Nxy, m=m, n=n)

    return plate1.buckling_analysis()[0][0]


def normalize_critical_load(Nxx, Nyy, Nxy):
    '''Normalizes the critical load to be used in optimization functions'''
    max_abs_load = [abs(Nxx), abs(Nyy), abs(Nxy)]
    max_abs_load.sort()
    max_abs_load = max_abs_load[-1]

    Nxx_norm = Nxx / max_abs_load
    Nyy_norm = Nyy / max_abs_load
    Nxy_norm = Nxy / max_abs_load

    return Nxx_norm, Nyy_norm, Nxy_norm, max_abs_load


def penalty_g1(y0):
    '''Defines the boundary conditions g1 that delimits the triangle.'''

    if len(y0) == 2: #unpacking for maximize_buckling
        xi1, xi3 = y0
    elif len(y0) == 3: #unpacking for minimize_panel_weight
        T, xi1, xi3 = y0

    g1 = xi3 + 2*xi1 + 1
    return g1


def penalty_g2(y0):
    '''Defines the boundary conditions g2 that delimits the triangle.'''
    if len(y0) == 2: #unpacking for maximize_buckling
        xi1, xi3 = y0
    elif len(y0) == 3: #unpacking for minimize_panel_weight
        T, xi1, xi3 = y0

    g2 = xi3 - 2*xi1 + 1
    return g2


def natural_constraint_g(y0):
    '''Defines the boundary conditions g1 that delimits the parabola.'''

    if len(y0) == 2: #unpacking for maximize_buckling
        xi1, xi3 = y0
    elif len(y0) == 3: #unpacking for minimize_panel_weight
        T, xi1, xi3 = y0

    g = xi3 - 2*xi1**2 + 1
    return g


def check_loads(Nxx, Nyy, Nxy):
    '''Given the load inputs, check if the function may present a problem'''

    # Positive loads warning   
    if Nxx >= 0 and Nyy >= 0 and Nxy == 0:
        warnings.warn(f'Buckling analysis is supposed to take at least a negative normal load or shear. (Nxx = {Nxx}, Nyy = {Nyy}, Nxy = {Nxy})')

    #Non normalized loads warning
    if abs(Nxx) > 1 or abs(Nyy) > 1 or abs(Nxy) > 1:
        warnings.warn(f'Loads will be normalized, prefer to use loads between -1 and 1 (Nxx = {Nxx}, Nyy = {Nyy}, Nxy = {Nxy})')


def _constraint(xi1, xi3, silent=True, penalty=True):
    g_natural = xi3 - 2*xi1**2 + 1
    g1 = xi3 + 2*xi1 + 1
    g2 = xi3 - 2*xi1 + 1

    if not penalty:
        g = g_natural
    elif xi1 < 0:
        g = g1
    else:
        g = g2

    if g < 0:
        if not silent:
            print(xi1, xi3)
        return False
    else:
        return True



def plot_optimization(a, b, T, m, n, E1, E2, v12,
                       G12, Nxx, Nyy, Nxy, panel_constraint, points_to_plot, res, penalty):

        print('generating plot...')
        x = np.linspace(-1., 1., points_to_plot)
        init_args = [0, 0, -1]
        Nx_arr = []
        xi1_arr, xi3_arr = [], []

        for xi1_ in x:
            for xi3_ in x:
                g_curr = _constraint(xi1_, xi3_, silent=True, penalty=penalty)
                if g_curr:
                    critc_N = Ncr_from_lp(a, b, T, m, n, xi1_, xi3_, E1, E2, v12, G12, Nxx, Nyy, Nxy, panel_constraint)
                    Nx_arr.append(critc_N)
                    xi1_arr.append(xi1_)
                    xi3_arr.append(xi3_)

                    if critc_N > init_args[2]:
                        init_args = [xi1_, xi3_, critc_N]
                else:
                    pass
                    Nx_arr.append(np.nan)
                    xi1_arr.append(xi1_)
                    xi3_arr.append(xi3_)

        g1_xi1 = np.linspace(-1, 0, points_to_plot)
        g2_xi1 = np.linspace(0, 1, points_to_plot)

        if penalty:
            g1_xi3 = -2*g1_xi1 - 1
            g2_xi3 = 2*g2_xi1 - 1
        else:
            print('penalty False')
            g1_xi3 = 2*g1_xi1**2 - 1
            g2_xi3 = 2*g2_xi1**2 - 1


        n_valid_points = int(np.sqrt(len(Nx_arr)))
        Nx_arr = np.array(Nx_arr)
        xi1_arr = np.array(xi1_arr)
        xi3_arr = np.array(xi3_arr)                       
        Nx_arr = Nx_arr.reshape(n_valid_points, n_valid_points)
        xi1_arr = xi1_arr.reshape(n_valid_points, n_valid_points)
        xi3_arr = xi3_arr.reshape(n_valid_points, n_valid_points)

        #plots
        plt.figure()

        # Nxcrit
        cs1 = plt.contour(xi1_arr, xi3_arr, Nx_arr, 20)
        plt.clabel(cs1)

        #g1 and g2
        plt.plot(g1_xi1, g1_xi3, 'k')
        plt.plot(g2_xi1, g2_xi3, 'k')

        #
        plt.plot(*res['x'], 'ro', label='optimum')


        plt.xlabel('xi1')
        plt.ylabel('xi3')
        plt.legend()
        plt.xlim([-1, 1])
        plt.ylim([-1, 1])
        plt.show()


'''
Test of PlateStructure.buckling_analysis
========================================

Mathematical verification performed according compmech.
'''

import pytest
import numpy as np

from math import isclose

from composipy import OrthotropicMaterial, LaminateProperty, PlateStructure


@pytest.fixture
def plateconfig():
    #Properties (N and mm)
    E1 = 60800
    E2 = 58250
    v12 = 0.07
    G12 = 4550
    t = 0.21

    # Plate Parameters
    a = 360
    b = 360
    m = 10
    n = 10

    stacking = [45,-45,0,90,45,-45,0,90,45,-45]
    stacking += stacking[::-1]

    a = 360.
    b = 360.

    mat1 = OrthotropicMaterial(E1, E2, v12, G12, t)
    laminate1 = LaminateProperty(stacking, mat1)

    plateconfig = {
        'dproperty' : laminate1,
        'a': a, 
        'b': b,
        'Nxx': -1, 
        'Nyy': -1,
        'm': m,
        'n': n
    }
   
    return plateconfig


def test_clamped_plate(plateconfig):
    first_eigen_reference = 119.76079572 # result from compmech
    plate1 = PlateStructure(**plateconfig, constraints='CLAMPED')
    eigenvalues, eigvectors = plate1.buckling_analysis()
    assert isclose(eigenvalues[0], first_eigen_reference, abs_tol=1e-4, rel_tol=0.1)


def test_pinned_plate(plateconfig):
    first_eigen_reference = 48.52457849 # result from compmech: p1.w1rx = p1.w2rx = p1.w1ry = p1.w2ry = 1
    plate1 = PlateStructure(**plateconfig, constraints='PINNED')
    eigenvalues, eigvectors = plate1.buckling_analysis()
    assert isclose(eigenvalues[0], first_eigen_reference, abs_tol=1e-4, rel_tol=0.1)